#include "algo.h"
extern int X,Y; 
extern Graph G;
extern double beta;
extern double deltabeta;
extern int nsweeps;

struct val_index * vpart;
edgepart part;

#define sign(a)   (((a)>0)?1:-1)

int f11=0,f12=0,f21=0,f22=0;
double *W11;
double *W12;
double *W21;
double *W22;
int *L11;
int *L12;
int *L21;
int *L22;

void  EdgePartitions(Graph , int  );
int RobinsonFoulds(Graph , char *** , int * );
void ReturnLinkNeighbors( Graph , int , int * , int *, int *, int *, int * , int *, int * , int *, int *, int *);
void ComputePauplin(Graph , int , int , double , double * , double * , int * , int *);
void PrintNewickFormat( Graph , int ** , FILE* ); 
void Compute_Leaves_Dist_Topological ( Graph g , int ** d);
double CalcolaLengthPauplin(Graph g, int **d);
int Compatibility();
void build_family(Graph , int , int, int *,int * );


double CalcolaEnergiaLink( int e, int s1, int s2, int s11, int s12, int s21, int s22,  int e11, int e12, int e21, int e22 )
{

	int i,j,s,k;
	double dist;
	double dA, dB, dC,E,dij,dks,dik,dis,djk,djs;


	f11=0; f12=0; f21=0; f22=0;
	ComputePauplin(G , s11, e11 , 1 , &dist, W11 , L11, &f11);
	ComputePauplin(G , s12, e12 , 1 , &dist, W12 , L12, &f12);
	ComputePauplin(G , s21, e21 , 1 , &dist, W21 , L21, &f21);
	ComputePauplin(G , s22, e22,  1 , &dist, W22 , L22, &f22);


	E=0;
	for( i = 0; i<f11; ++i ){
	    for( j = 0; j<f12; ++j ){
		for( k = 0; k<f21; ++k ){
		    for( s = 0; s<f22; ++s ){
			dij=  G.leavesdist[L11[i]][L12[j]];
			dks=  G.leavesdist[L21[k]][L22[s]];
			dik=  G.leavesdist[L11[i]][L21[k]];
			djs=  G.leavesdist[L12[j]][L22[s]];
			dis=  G.leavesdist[L11[i]][L22[s]];
			djk=  G.leavesdist[L12[j]][L21[k]];
			dA = dij + dks;
			dB = dik + djs;
			dC = dis + djk;
			E += max(0, (dA - min(dB,dC))/(dA + min(dB,dC))/(dA + min(dB,dC))/(dA + min(dB,dC))/(dA + min(dB,dC))/(dA + min(dB,dC)))*1.0 / W11[i] /  W12[j]   /W21[k] /  W22[s]  ;


		    }
		}
	    }
	}



	return E;

}


double CalcolaEnergiaTot( Graph g , int *countzero)
{

	int i;
	double E=0,Eneloc;
	int edge;
	int s1, s2, s11, s12, s21, s22;
	int e11,e12,e21,e22;
	
	*countzero=0;
	
	for(i=0; i < g.internaledge.size ;i++){

		edge = g.internaledge.vec[i];
		ReturnLinkNeighbors(g, edge, &s1, &s2, &s11, &s12, &s21, &s22,&e11,&e12,&e21,&e22);
		Eneloc=CalcolaEnergiaLink(edge, s1, s2,  s11, s12, s21, s22,e11,e12,e21,e22);
	
		E+=Eneloc;
		*countzero+=(Eneloc==0);


	}

	return E;
}

void Reshape( int e, int s1, int s2, int s11, int s12, int s21, int s22,  int e11, int e12, int e21, int e22 ,int *cswi)
{

	int c,i,j,k,s;
	double dist;
	double dA, dB, dC;
	double r,E = 0,EIN;
        double dij,dks,dik,djs,dis,djk;



	f11=0; f12=0; f21=0; f22=0;
	ComputePauplin(G , s11, e11 , 1 , &dist, W11 , L11, &f11);
	ComputePauplin(G , s12, e12 , 1 , &dist, W12 , L12, &f12);
	ComputePauplin(G , s21, e21 , 1 , &dist, W21 , L21, &f21);
	ComputePauplin(G , s22, e22,  1 , &dist, W22 , L22, &f22);

	c=2;


	EIN=CalcolaEnergiaLink(e, s1, s2,  s11, s12, s21, s22,e11,e12,e21,e22);
	c = ( int ) ( FRANDOM * 2 );	
	switch (c) {
	case 0:
	    E=0;
	    for( i = 0; i<f11; ++i ){
		for( j = 0; j<f21; ++j ){
		    for( k = 0; k<f12; ++k ){
			for( s = 0; s<f22; ++s ){
			    dij=  G.leavesdist[L11[i]][L21[j]];
			    dks=  G.leavesdist[L12[k]][L22[s]];
			    dik=  G.leavesdist[L11[i]][L12[k]];
			    djs=  G.leavesdist[L21[j]][L22[s]];
			    dis=  G.leavesdist[L11[i]][L22[s]];
			    djk=  G.leavesdist[L21[j]][L12[k]];
			    dA = dij + dks;
			    dB = dik + djs;
			    dC = dis + djk;
			    E += max(0, (dA - min(dB,dC))/(dA + min(dB,dC))/(dA + min(dB,dC))/(dA + min(dB,dC))/(dA + min(dB,dC))/(dA + min(dB,dC)))*1.0 /  W11[i] /  W21[j]   /  W12[k] /  W22[s]  ;
			}
		    }
		}
	    }
	    r = FRANDOM;

	    if (r > exp(-beta * sign(E-EIN)))
		c = 2;
		break;
	case 1:
	    E=0;
	    for( i = 0; i<f11; ++i ){
		for( j = 0; j<f22; ++j ){
		    for( k = 0; k<f12; ++k ){
			for( s = 0; s<f21; ++s ){
			    dij=  G.leavesdist[L11[i]][L22[j]];
			    dks=  G.leavesdist[L12[k]][L21[s]];
			    dik=  G.leavesdist[L11[i]][L12[k]];
			    djs=  G.leavesdist[L22[j]][L21[s]];
			    dis=  G.leavesdist[L11[i]][L21[s]];
			    djk=  G.leavesdist[L22[j]][L12[k]];
			    dA = dij + dks;
			    dB = dik + djs;
			    dC = dis + djk;
			    E += max(0, (dA - min(dB,dC))/(dA + min(dB,dC))/(dA + min(dB,dC))/(dA + min(dB,dC))/(dA + min(dB,dC))/(dA + min(dB,dC)))*1.0 /  W11[i] /  W22[j]   / W12[k] / W21[s]  ;
			}
		    }
		}
	    }
	    
	    r = FRANDOM;

	    if (r > exp(-beta * sign(E-EIN)))
		c = 2;
	    break;
	    default:	
		fprintf(stderr,"only two cases (0,1)\n");
		exit(1);
	}


	*cswi = c;	

	switch ( c ) {
	case 0:
		G.Edge[e21].s1 = s1;
		G.Edge[e21].s2 = s21;
		G.Edge[e12].s1 = s2;
		G.Edge[e12].s2 = s12;
		G.Node[s1].neigh[0] = e;
		G.Node[s1].neigh[1] = e11;
		G.Node[s1].neigh[2] = e21;
		G.Node[s2].neigh[0] = e;
		G.Node[s2].neigh[1] = e12;
		G.Node[s2].neigh[2] = e22;
		break;
	case 1: 
		G.Edge[e22].s1 = s1;
		G.Edge[e22].s2 = s22;
		G.Edge[e12].s1 = s2;
		G.Edge[e12].s2 = s12;
		G.Node[s1].neigh[0] = e;
		G.Node[s1].neigh[1] = e11;
		G.Node[s1].neigh[2] = e22;
		G.Node[s2].neigh[0] = e;
		G.Node[s2].neigh[1] = e12;
		G.Node[s2].neigh[2] = e21;		
		break;
	case 2:
		break;
			
	default:
		fprintf(stderr,"only three cases (0,1,2)\n");
		exit(1);
	}


}



void find_cherries( int s11, int e11) 
{

	int i;


	
        
for( i = 0; i< G.nnodes; ++i ){
		G.Node[i].prob=0.;
                G.Node[i].weights[0]=1.;
                G.Node[i].weights[1]=1.;
		G.Node[i].sons[0]=-1;
                G.Node[i].sons[1]=-1;
		
	}

    f11=0;
	build_family(G,s11,e11,L11,&f11);

       
        

 

if(f11==2){
X=L11[0];
Y=L11[1];
}


}
/*************************************************/

void permut(int num, int *perm){

	int count,index,temp;
	count=num-2;
	while (count>=0){
		index=FRANDOM*count;
		temp=perm[count+1];
		perm[count+1]=perm[index];
		perm[index]=temp;
		count--;
	} 


}


void Mc( char *nomegrafo)
{
	int i,j, ** disttopo,ciclo,*perm,countc,index;
	double E,minE;
	int edge,c;
	int s1, s2, s11, s12, s21, s22;
	int e11,e12,e21,e22;
	FILE *fpgrafo;
	int countzero,maxcountzero;


	W11 = (double *) malloc (G.nleaves * sizeof ( double ) );
	W12 = (double *) malloc (G.nleaves * sizeof ( double ) );
	W21 = (double *) malloc (G.nleaves * sizeof ( double ) );
	W22 = (double *) malloc (G.nleaves * sizeof ( double ) );

	L11 = (int *) malloc (G.nleaves * sizeof ( int ) );
	L12 = (int *) malloc (G.nleaves * sizeof ( int ) );
	L21 = (int *) malloc (G.nleaves * sizeof ( int ) );
	L22 = (int *) malloc (G.nleaves * sizeof ( int ) );

        perm = (int *) malloc (G.internaledge.size * sizeof ( int ) );
	for( i =0; i<G.internaledge.size; ++i){
		perm[i]=i;
	}

     
   fpgrafo = fopen ( nomegrafo, "w" ) ;

        if ( fpgrafo == NULL ) {
                fprintf(stderr, "grafo output file  %s doesn't exist ... exiting\n", nomegrafo );
                exit(-1);
        }



     

	disttopo = (int ** ) malloc ( G.nleaves * sizeof( int * ) );


	for( i = 0; i< G.nleaves; ++i ){
		disttopo[i] = (int *) malloc ( G.nleaves * sizeof( int ));
	}


	for( i=G.nleaves; i< G.nnodes ; i ++){
		G.Node[i].labelnewick = (char * ) malloc (10000 * sizeof ( char ) );
	}

	E=CalcolaEnergiaTot(G ,&countzero);

	beta = 0;
	minE=100000;
	maxcountzero=0;


	
	for( j = 0 ; j<10000; ++j ){
		

		countc=0;
		for(ciclo=0;ciclo<nsweeps;ciclo++){
		  for( i =0; i<G.internaledge.size; ++i){
		    
		    index=perm[i];
			edge = G.internaledge.vec[index];
			ReturnLinkNeighbors( G,edge, &s1, &s2, &s11, &s12, &s21, &s22,&e11,&e12,&e21,&e22);
			
			Reshape(edge, s1, s2,  s11, s12, s21, s22,e11,e12,e21,e22,&c);
			if(c!=2) countc++;
		}
		}
	
		if(!countc) break;
		permut(G.internaledge.size,perm);
							 beta+=deltabeta;
			}
  
for( i =0; i<G.internaledge.size; ++i){
			
			index=perm[i];
			edge = G.internaledge.vec[index];
			ReturnLinkNeighbors( G,edge, &s1, &s2, &s11, &s12, &s21, &s22,&e11,&e12,&e21,&e22);
			
			find_cherries(s11, e11);
                        find_cherries(s12, e12);
                        find_cherries(s21, e21);
                        find_cherries(s22, e22);
			
		       }

 	PrintNewickFormat( G,disttopo,fpgrafo ); 
	free( W11);
	free( W12);
	free( W21);
	free( W22);
	free( L11);
	free( L12);
	free( L21);
	free( L22);
	for( i = 0; i< G.nleaves; ++i ){
		free(disttopo[i]);
	}
	free(disttopo);


}
